# GDB 'Cilk support' command.
# Copyright (C) 2012 Free Software Foundation, Inc.

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import gdb


class CilkUtils(object):

    """Internal class which provides utilities for the main command classes."""

    @staticmethod
    def getCilkVersion():
        version = gdb.parse_and_eval("(((int*)(*((long long int*) &__cilkrts_global_state )))[0] >> 16) & 0xFFFF")
        return version or 0

    @staticmethod
    def valueFromInput(arg):
        on = ["on", "1"]
        off = ["off", "0"]
        value = arg.lower()
        if value in on:
            return 1
        if value in off:
            return 0
        else:
            raise gdb.GdbError("ERROR: This command requires on/off or 1/0 as argument.")

    @staticmethod
    def displaySerializationState():
        state = None
        if not CilkUtils.getCilkVersion():
            state = gdb.parse_and_eval(
                " ((int*)(*((long long int*) &__cilkrts_global_state )))[5]")
        if state is None:
            gdb.GdbError("Error: Cilk symbol not found.")
        print 'On' if state else 'Off'

    @staticmethod
    def setCilkSerialization(state):
        if not CilkUtils.getCilkVersion():
            gdb.parse_and_eval(
                " ((int*)(*((long long int*) &__cilkrts_global_state )))[5] = %s" % state)


class CilkSetSerialCommand(gdb.Command):

    """Set the Cilk runtime serialization state.

       Usage:

         set cilk-serialization STATE

         - STATE \"on\" or \"off\".
           On:  Turns Cilk serialization on.
           Off: Turns serialization off.
    """

    def __init__(self, state):
        super(
            CilkSetSerialCommand, self).__init__(name="set cilk-serialization",
                                                 command_class=gdb.COMMAND_DATA,
                                                 prefix=True)

    def invoke(self, arg_str, from_tty):
        global initialized
        if not initialized:
            print "Cilk run time not present yet."
            return
        try:
            state = CilkUtils.valueFromInput(arg_str)
        except:
            print "Please enter \"on\" or \"off\"."
            return
        self.setCilkSerialization(state)

    def setCilkSerialization(self, state):
        global cilkSerializationState
        CilkUtils.setCilkSerialization(state)
        cilkSerializationState = state
        CilkUtils.displaySerializationState()


class CilkShowSerializationCommand(gdb.Command):

    """Show state of Cilk serialization.
       Output displays the serialization state after setting.

       Usage:

         show cilk-serialization

         No argument.
         States are:
         On: Serialization is turned on.
         Off: Serialization is turned off.
    """

    def __init__(self):
        super(
            CilkShowSerializationCommand, self).__init__(name="show cilk-serialization",
                                                         command_class=gdb.COMMAND_DATA,
                                                         prefix=True)

    def invoke(self, arg_str, from_tty):
        global initialized
        if not initialized:
            print "Cilk run time not present yet."
            return
        self.showCilkSerialization()

    def showCilkSerialization(self):
        state = CilkUtils.displaySerializationState()


def lookup(symbol):
    debuggee = gdb.selected_inferior()
    if debuggee:
        try:
            lookedSymbol = gdb.parse_and_eval(symbol)
        except:
            return False
    return lookedSymbol


def exit_handler(event):
    global initialized, cilkSerialization, cilkShow
    cilkSerialization = None
    cilkShow = None
    initialized = False


def new_objfile_handler(event):
    global initialized, cilkSerialization, CilkShow
    if not initialized and lookup("__cilkrts_global_state"):
        if not cilkSerialization:
            cilkSerialization = CilkSetSerialCommand(
                cilkSerializationState)
            cilkShow = CilkShowSerializationCommand()
            initialized = True
        if cilkSerializationState:
            cilkSerialization.setCilkSerialization(cilkSerializationState)

cilkSerializationState = None
cilkSerialization = None
cilkShow = None
initialized = False
gdb.events.new_objfile.connect(new_objfile_handler)
gdb.events.exited.connect(exit_handler)
