# OpenMP support for GDB, the GNU debugger
#
# Copyright (C) 2015 Free Software Foundation, Inc.
#
# Contributed by Intel Corporation, <walfred.tedeschi@intel.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#

import string
import sys
import os
import struct

import gdb

import intel_openmp_dbg_hlp as rtl_helper
import openmp_commons


class IntelOpenMPRTL(openmp_commons.OpenMPRTL):
    debug_symbol = rtl_helper.debug_symbol

    def __init__(self, inferior):
        self.omp_threads_var = None
        self.cache_is_valid = 0
        self.omp_threads_var = None
        self.gdb_threads_var = {}
        self.ompId_to_gdbId = {}
        super(IntelOpenMPRTL, self).__init__(inferior)

    def check_size(self):
        return rtl_helper.check_size_of_debug_structure()

    @property
    def locks(self):
        self.init_threads()
        locks = rtl_helper.get_locks()
        if not locks:
            return
        for value in locks.itervalues():
            value.owner = self.get_gdbid_from_ompid(value.owner - 1)
        return locks

    def set_parent_and_incomplete_on_tasks(self, tasks):
        for id_, task in tasks.iteritems():
            parent_id = task.parent_id
            if parent_id in tasks:
                parent = tasks[parent_id]
                parent.append_child(id_)
                if parent_id in tasks:
                    parent = tasks[parent_id]
                    parent.append_child(task.id_)
                    if not task.is_complete:
                        parent.append_unfinished(task.id_)

    @property
    def tasks(self):
        self.init_threads()
        self._tasks = {}
        for thread in self.omp_threads_var:
            # get current running tasks...
            rtl_helper.get_task_and_parents(
                self._tasks, self.omp_threads_var[thread].task_address)
            # get team tasks
            rtl_helper.get_task_team(
                self._tasks, self.omp_threads_var[thread].task_team_address_0)
            rtl_helper.get_task_team(
                self._tasks, self.omp_threads_var[thread].task_team_address_1)
        if self._tasks:
            self.set_parent_and_incomplete_on_tasks(self._tasks)
        return self._tasks

    def init_threads(self):
        if not self.omp_threads_var or not self.cache_is_valid:
            self.gdb_threads
            self.omp_threads_var = rtl_helper.get_threads(self.gdb_threads_var)

    @property
    def threads(self):
        self.init_threads()
        return self.omp_threads_var

    @property
    def team(self):
        self.init_threads()
        teams = {}
        root_team = rtl_helper.get_root_team()
        if root_team:
            gdb_id_of_master = self.get_gdbid_from_ompid(root_team.master)
            root_team.set_master_gdb_id(gdb_id_of_master)
            teams[root_team.team_id] = root_team
        for thread in self.omp_threads_var:
            if self.omp_threads_var[thread].task_address:
                team1 = rtl_helper.get_team_info(
                    self.omp_threads_var[thread].team_address)
            if not team1.team_id in teams:
                gdb_id_of_master = self.get_gdbid_from_ompid(team1.master)
                team1.set_master_gdb_id(gdb_id_of_master)
                teams[team1.team_id] = team1
            teams[team1.team_id].append_member(
                self.omp_threads_var[thread].gdb_id)
        return teams

    def test_barrier_set_threads(self, barriers, current_barrier, team):
        if current_barrier.workers_arrived:
            barriers[current_barrier.id_] = current_barrier
            threads = team.members
            for index in threads:
                address = self.get_ompthread_from_gdbid(index).barrier_base
                if rtl_helper.get_thread_has_arrived(address, current_barrier.kind):
                    current_barrier.append_threads_arrived(index)

    @property
    def dependency(self):
        return {index: task.dependency for index, task in self.tasks.iteritems() if task.dependency}

    @property
    def barriers(self):
        teams = self.team
        if not teams:
            return
        barriers = {}
        for team in teams.itervalues():
            self.test_barrier_set_threads(
                barriers, team.barrier_explicit, team)
            self.test_barrier_set_threads(
                barriers, team.barrier_implicit, team)
            self.test_barrier_set_threads(
                barriers, team.barrier_reduction, team)
        return barriers

    def get_gdbid_from_ompid(self, val):
        # takes into account tha self.threads and self.gdb_threads are already
        # known.
        for thread in self.omp_threads_var:
            if self.omp_threads_var[thread].omp_id == val:
                return self.omp_threads_var[thread].gdb_id

    def get_ompthread_from_gdbid(self, val):
        # takes into account tha self.threads and self.gdb_threads are already
        # known.
        for item in self.omp_threads_var.itervalues():
            if item.gdb_id == val:
                return item

    @property
    def gdb_threads(self):
        # if not self.cache_is_valid:
        threads = gdb.selected_inferior().threads()
        for thread in threads:
            handler_num = rtl_helper.get_thread_handler(thread.num)
            if handler_num:
                self.gdb_threads_var[handler_num] = thread.num
        return self.gdb_threads_var

    @property
    def taskwaits(self):
        taskwaits = {}
        tasks = self.tasks
        for id_, task in tasks.iteritems():
            taskwait = task.taskwait
            if taskwait.counter:
                taskwaits[taskwait.id_] = taskwait
                # Till the thread ID reported on taskwait is the OMPRTL
                # Fixing it.
                taskwait.thread = self.get_gdbid_from_ompid(taskwait.thread + 1)
                taskwait.set_waiting_tasks(task.children_nums)
        return taskwaits

    def invalidate_cache(self):
        self.cache_is_valid = False

    def has_supported_version(self):
        version = gdb.parse_and_eval(rtl_helper.version_debug_str)
        return version == 9 and self.check_size()


def set_kmp_debug(event):
    try:
        gdb.parse_and_eval('__kmp_debugging = 1\n')
    except:
        pass

gdb.events.new_objfile.connect(set_kmp_debug)
