# OpenMP support for GDB, the GNU debugger
#
# Copyright (C) 2015 Free Software Foundation, Inc.
#
# Contributed by Intel Corporation, <walfred.tedeschi@intel.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
import gdb


class OMPException(Exception):
    _default_message = 'BASE OMPException raised'

    def __init__(self, msg=''):
        self.args = [msg or self._default_message]


class OMPRTLNotRecognized(OMPException):
    _default_message = 'OpenMP RTL unsuported.'


class OMPNoInferior(OMPException):
    _default_message = 'Inferior not present yet.'


class OMPNotSupportedSize(OMPException):
    _default_message = 'Size not supported to read integers.'


class OMPMoreLocksThanExpected(OMPException):
    _default_message = 'Found discrepancy on locks used/allocated.'


class OMPNotSupportedSize(OMPException):
    _default_message = 'Maxsize too small for the string.'

class OMPNotSupportedInput(OMPException):
    _default_message = 'Maxsize too small for the string.'

# fix-me: names must be changed
class Thread(object):

    def __init__(self, handler, omp_tid, omp_global_tid, gdb_tid):
        self.handler, self.omp_tid, self.omp_gtid, self.gdb_id = hdl, o_tid, o_gtid, gdb_id

    def __str__(self):
        return 't_handler = 0x%x: omp_tid = 0x%x, omp_gtid = 0x%x gdb_id = 0x%x' % (self.handler, self.omp_tid, self.omp_gtid, self.gdb_id)


class OpenMPRTL(object):
    debug_symbol = ''

    def __init__(self, inferior):
        self.rtl_symbol = gdb.lookup_global_symbol(self.debug_symbol)
        self.inferior = inferior

    @property
    def locks(self):
        return

    @property
    def tasks(self):
        return

    def has_supported_version(self):
        return False


class Lock(object):

    def __init__(self, lock_id, location, kind, status, depth, owner):
        self.lock_id = lock_id
        self.kind = kind
        self.status = status
        self.depth = depth
        self.location = location
        self.owner = owner

    @staticmethod
    def print_header():
        print 'Id\tStatus\tkind\tDepth\tOwner\tLocation'

    def __str__(self):
        return '%d\t%s\t%s\t%s\t%s\t%s' % (self.lock_id, self.status, self.kind, self.depth, self.owner or '---', self.location)


class Routine(object):

    def __init__(self, address, function, compilation_file, line):
        self.address = address
        self.function = function or '---'
        self.location = (
            '%s:%d' % (compilation_file, line)) if compilation_file and line else '---'

    def __str__(self):
        return '0x%x %s %s' % (self.address, self.function, self.location)


class DependentTaskNode(object):

    def __init__(self, npredecessors, routine):
        self.npredecessors = npredecessors
        self.routine = routine
        self.successors = []
        self.node = []

    def append_node(self, new_node):
        self.successors.append(new_node)

    def __str__(self):
        items = ['\t[%d]\t%s' % (self.npredecessors, self.routine)] + \
            ['\t\t[%d]\t%s' % (successor.npredecessors, successor.routine)
             for successor in self.successors]
        return '\n'.join(items)


class TaskDependency(object):

    def __init__(self, id_):
        self.id_ = id_
        self.dependent_tasks = []

    def append_dependent_task(self, task):
        self.dependent_tasks.append(task)

    @staticmethod
    def print_header():
        print 'Task id.\t\tSuccessor tasks (ordered)'

    def __str__(self):
        if self.dependent_tasks:
            return '%d%s\n' % (self.id_,
                               ''.join('\t%s\n' % task for task in self.dependent_tasks))
        else:
            return ''

class Thread(object):

    def __init__(self, rtl_id, handler, tid, gdb_id):
        self.omp_id = rtl_id
        self.handler = handler
        self.tid = tid
        self.gdb_id = gdb_id

    @staticmethod
    def print_header():
        print 'id\tOpenMP id\tHandler\tThread id'

    def __str__(self):
        return '%x\t0x%x\t%x' % (self.omp_id, self.handler, self.tid)


class Taskwait(object):

    def __init__(self, taskwait_id, task, location, counter, thread):
        self.waiting_tasks = []
        self.id_ = taskwait_id
        self.task_id = task
        self.thread = thread
        self.location = location
        self.counter = counter

    def set_waiting_tasks(self, tasks):
        self.waiting_tasks = tasks

    def append_waiting_tasks(self, id):
        self.waiting_tasks.append(id)

    def __str__(self):
        header = '%d\ttaskwait for task %d\n' % (self.id_ or -1, self.task_id or -1)

        thread = '\t\tEncontering thread: %s\n' % (str(self.thread).lower())
        creation = '\t\tCreated at: %s\n' % (self.location or 'unknown')
        waiting = '\t\tAwaiting tasks: %s\n' % (
                                                ', '.join(str(id_) for id_ in self.waiting_tasks) or 'none')
        return '%s%s%s%s' % (header, thread, creation, waiting)


class Barrier(object):

    def __init__(self, id_, kind, team, master_arrived, workers_arrived):
        self.id_ = id_
        self.kind = kind
        self.team = team
        self.master_arrived = master_arrived
        self.workers_arrived = workers_arrived
        self.threads_arrived = set()

    def append_threads_arrived(self, id_):
        self.threads_arrived.add(id_)

    def __str__(self):
        str_initial = '\n%d\t%s barrier on team: %d' % (
            self.id_, self.kind, self.team)
        str_master = '\t\tMaster has arrived: %s' % (
            'Yes' if self.master_arrived else 'No')
        str_threads_arrived = '\t\t%d workers arrived: %s' % (len(self.threads_arrived),
                                                              str(', '.join(str(s) for s in self.threads_arrived)) if self.threads_arrived else '---')
        return '%s\n%s\n%s\n' % (str_initial, str_master, str_threads_arrived)


class Task(object):

    def __init__(self, id_, flags, team, parent, level, ident, children, incomplete_children, status, is_complete):
        self.id_ = id_
        self.flags = flags
        self.team = team
        self.parent_id = parent
        self.level = level
        self.ident = ident
        self.children = children
        self.incomplete_children = incomplete_children
        self.status = status
        self.is_complete = is_complete
        self.children_nums = set()
        self.unfinished = set()

    def append_child(self, id_):
        self.children_nums.add(id_)

    def append_unfinished(self, id_):
        self.unfinished.add(id_)

    def __str__(self):
        if not self.team:
            return
        str_initial = '\n%d\t\t%s task on team: %d' % (
            self.id_, self.status, self.team)
        str_parent = '\t\tParent task: %s' % (self.parent_id or '---')
        str_created = '\t\tCreated at: %s' % self.ident
        str_child = '\t\tChildren tasks [%d]: %s' % (
            len(self.children_nums), ', '.join(str(id_) for id_ in self.children_nums) or '---')
        str_unfineshed = '\t\tUnfinished children [%d]: %s' % (len(self.unfinished),
            ', '.join(str(s) for s in self.unfinished) if self.unfinished else '---')
        return '%s\n%s\n%s\n%s\n%s' % (str_initial, str_parent, str_created, str_unfineshed, str_child)


class Team(object):

    def __init__(self, team_id, master_tid, parent, nproc, location):
        self.team_id = team_id
        self.master = master_tid
        self.master_gdb_id = -1
        self.parent = parent
        self.nproc = nproc
        self.location = location
        self.members = set()

    def append_member(self, gdb_id):
        if gdb_id > 0 and gdb_id not in self.members:
            self.members.add(gdb_id)

    def set_master_gdb_id(self, gdb_id):
        self.master_gdb_id = gdb_id

    def __str__(self):
        str_team = 'OpenMP Team:\t%d' % (self.team_id or -1)
        str_parent = 'Parent Team:\t%s' % (self.parent or '---')
        where_created = self.location or 'unknown'
        str_created = 'Created at:\t%s' % where_created
        str_master = 'Master Thread:\t[%d]' % (self.master_gdb_id or -1)
        str_members = 'Members:\n'
        members = "---"
        try:
            members = ('\t[%i] Thread %d\n' % (i + 1, member)
                       for i, member in enumerate(self.members))
        except:
            pass
        return '%s\n%s\n%s\n%s\n%s\n' % (str_team, str_master, str_parent, str_created, 'Members:\n' + ''.join(members))
