#
# GDB Python script for soft boundary check support.
#
# Copyright (c) 2012, Intel Corporation, <walfred.tedeschi@intel.com>.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms and conditions of the GNU General Public License,
# version 2, as published by the Free Software Foundation.
#
# This program is distributed in the hope it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin St - Fifth Floor, Boston, MA 02110-1301 USA.
#

# Caveat:
# Script is meant to be used on stop mode,
# i.e. default gdb mode. The python interface is
# taking into account gdb 7.4.
#
# Usage Notes:
# 1. to use the script you should source the script in the .gdbinit.
#    e.g. "source /$HOMEDIR/.gdb_scripts/softbounds_support.py",
#    taking into account that the script is under .gdb_scripts directory.
#

import gdb
import struct

def getValueFromBuffer (pointerSize, bufferM, offset):
    fmt = None
    if   pointerSize == 8: fmt = "@q"
    elif pointerSize == 4: fmt = "@l"
    else: raise gdb.GdbError ("Unknown pointer size.")
    value  = struct.unpack_from (fmt, bufferM, offset)[0]
    return value

def getPointerSize ():
    return gdb.parse_and_eval ("sizeof (char*)\n")

def displayNoSWBoundaryCheckSupport ():
    print "No software bounds support available."

def displayNoBoundaryInfoAvailable ():
    print "No boundary information available."

def displaySymbolNotSuitable ():
    print "Given symbol is not suitable. Pointer or array expected."

def displaySymbolNotFound (symbol):
    print "Symbol not found: %s." % (symbol)

def displayPointerNotAllocated (symbol):
    print "No bound info \"%s\" is a null pointer." % (symbol)

def displayBoundaryViolationDetected (lower, upper, accessed, size, pointerSize):
    if not lower: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if not upper: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if not accessed: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if not size: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if accessed > upper: print "Upper bound violation."
    else: print "Lower bound violation."
    if pointerSize == 64:
        print "With bounds {0x%x, 0x%x} accessing at: 0x%x and size: %d." % \
        ( lower & 0xFFFFFFFFFFFFFFFF, upper & 0xFFFFFFFFFFFFFFFF, accessed & 0xFFFFFFFFFFFFFFFF, size)
    else:
        print "With bounds {0x%x, 0x%x} accessing at: 0x%x and size: %d." % \
        ( lower & 0xFFFFFFFF, upper & 0xFFFFFFFF, accessed & 0xFFFFFFFF, size)

def displayBoundaryInfo (address, low, up, pointerSize):
    if not address: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if not low: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if not up: raise gdb.GdbError ("Wrong argument for display boundaries.")
    if pointerSize == 64:
        print "{lbound = 0x%x, ubound = 0x%x}: ptr. value = 0x%x, size = %d" % \
        (low & 0xFFFFFFFFFFFFFFFF, up & 0xFFFFFFFFFFFFFFFF, address & 0xFFFFFFFFFFFFFFFF, up - low + 1)
    else:
        print "{lbound = 0x%x, ubound = 0x%x}: ptr. value = 0x%x, size = %d" % \
        (low & 0xFFFFFFFF, up & 0xFFFFFFFF, address & 0xFFFFFFFF, up - low + 1)

def displayNotLinkAgainstPthread (error):
    print "Error: %s.\nTo use boundary violation detection please link application with pthread." % (error)

#
# Class Descriptor: Describes the boundaries of a pointer.
#
class Descriptor:
    def __init__ (self, lower, upper, address):
        self.lowbounds   = lower
        self.upperbounds = upper
        self.address     = address

    @staticmethod
    def fromBuffer (bufferM, pointer_size):
        if not buffer: raise gdb.GdbError ("Wrong argument to build descriptor.")
        lower   = getValueFromBuffer (pointer_size, bufferM, 0)
        upper   = getValueFromBuffer (pointer_size, bufferM, pointer_size)
        address = getValueFromBuffer (pointer_size, bufferM, pointer_size * 2)
        return Descriptor (lower,upper, address)

    #
    # Information might be described on the debug information
    # We use here the compile convention implemented to describe
    # this value.
    #
    @staticmethod
    def lookupBoundsOnDebugInfo (symbol):
        lowerBound = "__pl_low_bound_"  + symbol + "\n"
        upperBound = "__pl_high_bound_" + symbol + "\n"
        try:
            lower   = gdb.parse_and_eval (lowerBound)
            upper   = gdb.parse_and_eval (upperBound)
            address = gdb.parse_and_eval (symbol)
        except:
            return None
        descriptor = Descriptor (lower,upper, address)
        return descriptor

    def printMe (self):
        displayBoundaryInfo (self.address, self.lowbounds, self.upperbounds, getPointerSize ())

#
# Class MapInfo: describes the actual data storage algorithm used to archive
# descriptors for pointer.
#
class MapInfo:
    def __init__ (self, pointer_size, inferior, pageBits, firstLevel, mapAddress, pageMask):
        self.pointerSize = pointer_size
        self.inferior   = inferior
        self.pageBits   = pageBits
        self.firstLevel = firstLevel
        self.mapAddress = mapAddress
        self.pageMask   = pageMask

    @staticmethod
    def createMapInfo (bufferM, pointer_size, inferior):
        if not buffer: raise gdb.GdbError ("Wrong argument to create Map info.")
        if not inferior: raise gdb.GdbError ("Wrong argument to create Map info.")
        pageBits   = getValueFromBuffer (pointer_size, bufferM, 0)
        firstLevel = getValueFromBuffer (pointer_size, bufferM, pointer_size)
        mapAddress = getValueFromBuffer (pointer_size, bufferM, pointer_size * 2)
        # pageSize is a temporary variable.
        pageSize   = 1 << pageBits
        pageMask   = ( ( pageSize - 1 ) & ~( pointer_size - 1 ))
        return MapInfo (pointer_size, inferior, pageBits, firstLevel, mapAddress, pageMask)

    def getDescriptor (self, addressValue):
        address = long (addressValue)
        if address == 0: return None
        key = address >> self.pageBits
        descriptorAddr = self.mapAddress + (self.pointerSize * key)
        try:
            memory = self.inferior.read_memory (descriptorAddr, self.pointerSize)
        except:
            return None
        if not memory: return None
        deindirection  = getValueFromBuffer (self.pointerSize, memory, 0)
        nOfAttributesInC_Desc = 4
        if self.pointerSize == 64:
            descriptorLoc  = long (deindirection) + long ( ( (address & self.pageMask) / self.pointerSize) * (nOfAttributesInC_Desc * self.pointerSize))
        else:
            descriptorLoc  = (deindirection) + ( ( (address & self.pageMask) / self.pointerSize) * (nOfAttributesInC_Desc * self.pointerSize))
        try:
            memoryDesc = self.inferior.read_memory (descriptorLoc, nOfAttributesInC_Desc * self.pointerSize)
        except:
            return None
        descriptor = Descriptor.fromBuffer (memoryDesc, self.pointerSize)
        return descriptor

#
# Check for gdb api support.
#
softBoundCheckSupport=True;
try:
    accessLookupSymbol = gdb.lookup_symbol
    accessSymbolType   = gdb.Symbol.type
    accessFinishBP     = gdb.FinishBreakpoint
except AttributeError:
    softBoundCheckSupport=False

#
# Actual new command "show bounds"
#
if softBoundCheckSupport:
    class ShowSoftBounds (gdb.Command):
        """Displays the bounds for a given pointer or array: info sbounds POINTER/ARRAY
    Print "info sbounds VAR" where VAR is a pointer or an array."""

        def __init__ (self):
            gdb.Command.__init__ (self, "info sbounds", gdb.COMMAND_DATA)
            self.mapInfo = None
            self.pointerSize = 0

        def buildMapInfo (self):
            mapSymbol = gdb.parse_and_eval ("&__DBG_CHKP_BNDCFGU\n")
            if not mapSymbol: return
            self.pointerSize = getPointerSize ()
            if not self.pointerSize: return
            debuggee = gdb.selected_inferior ()
            if not debuggee: return
            derefFromBNDCFGU = debuggee.read_memory (mapSymbol, 1 * self.pointerSize)
            valueBNDCFGU = getValueFromBuffer (self.pointerSize, derefFromBNDCFGU, 0)
            memory = debuggee.read_memory (valueBNDCFGU, 3 * self.pointerSize)
            self.mapInfo = MapInfo.createMapInfo (memory, self.pointerSize, debuggee)

        def invoke (self, arg, from_tty):
            if not arg:
                raise gdb.GdbError ("\"info sbounds\" command takes one argument")
            hasSoftBoundsSupport = gdb.parse_and_eval ("__chkp_check_bounds\n")
            if not hasSoftBoundsSupport:
                displayNoSWBoundaryCheckSupport ()
                return
            value = gdb.parse_and_eval (arg)
            if value == None:
                displaySymbolNotFound (arg)
                return
            elif not value:
                displayPointerNotAllocated (arg)
                return
            if value.type.code != gdb.TYPE_CODE_PTR and value.type.code != gdb.TYPE_CODE_ARRAY:
                displaySymbolNotSuitable ()
                return
            descriptor = Descriptor.lookupBoundsOnDebugInfo (arg)
            if descriptor:
                descriptor.printMe ()
                return
            self.buildMapInfo ()
            if not self.mapInfo:
                displayNoBoundaryInfoAvailable ()
            descriptor = self.mapInfo.getDescriptor (value.address)
            if descriptor:
                descriptor.printMe ()
            else:
                displayNoBoundaryInfoAvailable ()

    #
    # FIXME: We should be able to get the information where the boundary
    # violation occurred from the run-time.
    #
    class BackToUser (gdb.FinishBreakpoint):
        def __init__ (self, frame, pc, address, low, up, size):
            gdb.FinishBreakpoint.__init__ (self, frame, internal=True)
            self.bp      = pc
            self.address = address
            self.low     = low
            self.up      = up
            self.size    = size

        def stop (self):
            displayBoundaryViolationDetected (self.address, self.low, self.up, self.size, getPointerSize ())
            return True


    class BoundaryViolationBP (gdb.Breakpoint):

            #
            # FIXME: The runtime should hid bnd_exc symbol with the prefix __
            #
            def __init__ (self):
                gdb.Breakpoint.__init__ (self, "bnd_exc", gdb.BP_BREAKPOINT, gdb.WP_ACCESS, internal=True)

            #
            # FIXME: The runtime should hide report_option symbol with the prefix __
            #
            def stop (self):
                gdb.execute ("set report_option = 0")
                debuggee = gdb.selected_inferior ()
                if not debuggee: raise gdb.GdbError ("No inferior running.")
                try:
                    bndExcSymbol = gdb.parse_and_eval ("&__DBG_CHKP_BNDEXCINFO\n")
                except gdb.error, e:
                    displayNotLinkAgainstPthread (e)
                    gdb.execute ("set report_option = 1")
                    return False
                if not bndExcSymbol: raise gdb.GdbError ("Boundary exception info symbol not present.")
                pointer_size = long (getPointerSize ())
                if not pointer_size: raise gdb.GdbError ("Pointer size could not be evaluated.")
                sizeof_boundaryException = 5
                memory   = debuggee.read_memory (bndExcSymbol, sizeof_boundaryException * pointer_size)
                pc       = getValueFromBuffer (pointer_size, memory, 0)
                lower    = getValueFromBuffer (pointer_size, memory, pointer_size)
                up       = getValueFromBuffer (pointer_size, memory, pointer_size * 2)
                accessed = getValueFromBuffer (pointer_size, memory, pointer_size * 3)
                size     = getValueFromBuffer (pointer_size, memory, pointer_size * 4)
                BackToUser (gdb.selected_frame ().older (), pc, lower, up, accessed, size)
                return False

    def lookup (symbol):
        debuggee = gdb.selected_inferior ()
        if debuggee:
            try:
                lookedSymbol = gdb.parse_and_eval (symbol)
            except:
                return False
        return lookedSymbol

    def new_objfile_handler (event):
        global initialized
        if not initialized and lookup ("__DBG_CHKP_BNDCFGU"):
                BoundaryViolationBP ()
                initialized = True

#
# Actual GDB initialization
#
    initialized = False
    gdb.events.new_objfile.connect (new_objfile_handler)
    ShowSoftBounds ()

